# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_basic_preset_query_with_presets 1'] = {
    'pipeline': {
        'name': 'csv_hello_world',
        'presets': [
            {
                '__typename': 'PipelinePreset',
                'environmentConfigYaml': '''solids:
  sum_solid:
    inputs:
      num: data/num_prod.csv
''',
                'mode': 'default',
                'name': 'prod',
                'solidSubset': None
            },
            {
                '__typename': 'PipelinePreset',
                'environmentConfigYaml': '''solids:
  sum_solid:
    inputs:
      num: data/num.csv
''',
                'mode': 'default',
                'name': 'test',
                'solidSubset': None
            }
        ]
    }
}



from __future__ import absolute_import

from dagster import (
    LoggerDefinition,
    ModeDefinition,
    PipelineDefinition,
    ResourceDefinition,
    PresetDefinition,
    check,
)
from dagster.core.definitions import (
    create_environment_schema,
    create_environment_type,
    IOExpectationDefinition,
)
from dagster_graphql import dauphin

from .config_types import to_dauphin_config_type
from .runtime_types import to_dauphin_runtime_type
from .solids import DauphinSolidContainer, build_dauphin_solid_handles, build_dauphin_solids


class DauphinPipeline(dauphin.ObjectType):
    class Meta:
        name = 'Pipeline'
        interfaces = [DauphinSolidContainer]

    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    solids = dauphin.non_null_list('Solid')
    environment_type = dauphin.Field(
        dauphin.NonNull('ConfigType'), mode=dauphin.String(required=False)
    )
    config_types = dauphin.Field(
        dauphin.non_null_list('ConfigType'), mode=dauphin.String(required=False)
    )
    runtime_types = dauphin.non_null_list('RuntimeType')
    runs = dauphin.non_null_list('PipelineRun')
    modes = dauphin.non_null_list('Mode')
    solid_handles = dauphin.non_null_list('SolidHandle')
    presets = dauphin.non_null_list('PipelinePreset')

    def __init__(self, pipeline):
        super(DauphinPipeline, self).__init__(name=pipeline.name, description=pipeline.description)
        self._pipeline = check.inst_param(pipeline, 'pipeline', PipelineDefinition)

    def resolve_solids(self, _graphene_info):
        return build_dauphin_solids(self._pipeline)

    def resolve_environment_type(self, _graphene_info, mode=None):
        return to_dauphin_config_type(create_environment_type(self._pipeline, mode))

    def resolve_config_types(self, _graphene_info, mode=None):
        environment_schema = create_environment_schema(self._pipeline, mode)
        return sorted(
            list(map(to_dauphin_config_type, environment_schema.all_config_types())),
            key=lambda config_type: config_type.key,
        )

    def resolve_runtime_types(self, _graphene_info):
        return sorted(
            list(
                map(
                    to_dauphin_runtime_type,
                    [t for t in self._pipeline.all_runtime_types() if t.name],
                )
            ),
            key=lambda config_type: config_type.name,
        )

    def resolve_runs(self, graphene_info):
        return [
            graphene_info.schema.type_named('PipelineRun')(r)
            for r in graphene_info.context.pipeline_runs.all_runs_for_pipeline(self._pipeline.name)
        ]

    def get_dagster_pipeline(self):
        return self._pipeline

    def get_type(self, _graphene_info, typeName):
        if self._pipeline.has_config_type(typeName):
            return to_dauphin_config_type(self._pipeline.config_type_named(typeName))
        elif self._pipeline.has_runtime_type(typeName):
            return to_dauphin_runtime_type(self._pipeline.runtime_type_named(typeName))

        else:
            check.failed('Not a config type or runtime type')

    def resolve_modes(self, graphene_info):
        return [
            graphene_info.schema.type_named('Mode')(mode_definition)
            for mode_definition in sorted(
                self._pipeline.mode_definitions, key=lambda item: item.name
            )
        ]

    def resolve_solid_handles(self, _graphene_info):
        return sorted(
            build_dauphin_solid_handles(self._pipeline), key=lambda item: str(item.handleID)
        )

    def resolve_presets(self, _graphene_info):
        return [
            DauphinPipelinePreset(preset, self._pipeline.name)
            for preset in sorted(self._pipeline.get_presets(), key=lambda item: item.name)
        ]


class DauphinPipelineConnection(dauphin.ObjectType):
    class Meta:
        name = 'PipelineConnection'

    nodes = dauphin.non_null_list('Pipeline')


class DauphinResource(dauphin.ObjectType):
    class Meta:
        name = 'Resource'

    def __init__(self, resource_name, resource):
        self.name = check.str_param(resource_name, 'resource_name')
        self._resource = check.inst_param(resource, 'resource', ResourceDefinition)
        self.description = resource.description

    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    configField = dauphin.Field('ConfigTypeField')

    def resolve_configField(self, graphene_info):
        return (
            graphene_info.schema.type_named('ConfigTypeField')(
                name="config", field=self._resource.config_field
            )
            if self._resource.config_field
            else None
        )


class DauphinLogger(dauphin.ObjectType):
    class Meta:
        name = 'Logger'

    def __init__(self, logger_name, logger):
        self.name = check.str_param(logger_name, 'logger_name')
        self._logger = check.inst_param(logger, 'logger', LoggerDefinition)
        self.description = logger.description

    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    configField = dauphin.Field('ConfigTypeField')

    def resolve_configField(self, graphene_info):
        return (
            graphene_info.schema.type_named('ConfigTypeField')(
                name="config", field=self._logger.config_field
            )
            if self._logger.config_field
            else None
        )


class DauphinMode(dauphin.ObjectType):
    def __init__(self, mode_definition):
        self._mode_definition = check.inst_param(mode_definition, 'mode_definition', ModeDefinition)

    class Meta:
        name = 'Mode'

    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    resources = dauphin.non_null_list('Resource')
    loggers = dauphin.non_null_list('Logger')

    def resolve_name(self, _graphene_info):
        return self._mode_definition.name

    def resolve_description(self, _graphene_info):
        return self._mode_definition.description

    def resolve_resources(self, _graphene_info):
        return [
            DauphinResource(*item) for item in sorted(self._mode_definition.resource_defs.items())
        ]

    def resolve_loggers(self, _graphene_info):
        return [DauphinLogger(*item) for item in sorted(self._mode_definition.loggers.items())]


class DauphinMetadataItemDefinition(dauphin.ObjectType):
    class Meta:
        name = 'MetadataItemDefinition'

    key = dauphin.NonNull(dauphin.String)
    value = dauphin.NonNull(dauphin.String)


class DauphinExpectation(dauphin.ObjectType):
    class Meta:
        name = 'Expectation'

    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()

    def __init__(self, expectation):
        check.inst_param(expectation, 'expectation', IOExpectationDefinition)
        super(DauphinExpectation, self).__init__(
            name=expectation.name, description=expectation.description
        )


class DauphinPipelinePreset(dauphin.ObjectType):
    class Meta:
        name = 'PipelinePreset'

    name = dauphin.NonNull(dauphin.String)
    solidSubset = dauphin.List(dauphin.NonNull(dauphin.String))
    environmentConfigYaml = dauphin.NonNull(dauphin.String)
    mode = dauphin.NonNull(dauphin.String)

    def __init__(self, preset, pipeline_name):
        self._preset = check.inst_param(preset, 'preset', PresetDefinition)
        self._pipeline_name = check.str_param(pipeline_name, 'pipeline_name')

    def resolve_name(self, _graphene_info):
        return self._preset.name

    def resolve_solidSubset(self, _graphene_info):
        return self._preset.solid_subset

    def resolve_environmentConfigYaml(self, _graphene_info):
        yaml = self._preset.get_environment_yaml(self._pipeline_name)
        return yaml if yaml else ''

    def resolve_mode(self, _graphene_info):
        return self._preset.mode



from __future__ import absolute_import
from collections import namedtuple
import copy
import logging
import os
import sys
import time
import atexit

import gevent
import six

from dagster import (
    ExecutionTargetHandle,
    InProcessExecutorConfig,
    PipelineDefinition,
    RunConfig,
    check,
    execute_pipeline,
)
from dagster.core.events.log import DagsterEventRecord
from dagster.core.events import DagsterEvent, DagsterEventType, PipelineProcessStartedData
from dagster.utils.error import serializable_error_info_from_exc_info, SerializableErrorInfo
from dagster.utils import get_multiprocessing_context
from dagster_graphql.implementation.pipeline_run_storage import PipelineRun


class PipelineExecutionManager(object):
    def execute_pipeline(self, handle, pipeline, pipeline_run, raise_on_error):
        raise NotImplementedError()


def build_synthetic_pipeline_error_record(run_id, error_info, pipeline_name):
    check.str_param(run_id, 'run_id')
    check.str_param(pipeline_name, 'pipeline_name')
    check.inst_param(error_info, 'error_info', SerializableErrorInfo)

    return DagsterEventRecord(
        message=error_info.message + '\nStack Trace:\n' + '\n'.join(error_info.stack),
        # Currently it is the user_message that is displayed to the user client side
        # in dagit even though that was not the original intent. The original
        # intent was that the user_message was the message generated by user code
        # communicated directly to the client. We need to rationalize the treatment
        # of these different error messages
        user_message=(
            'An exception was thrown during execution that is likely a framework error, '
            'rather than an error in user code.'
        )
        + '\nOriginal error message: '
        + error_info.message
        + '\nStack Trace:\n'
        + '\n'.join(error_info.stack),
        level=logging.ERROR,
        run_id=run_id,
        timestamp=time.time(),
        error_info=error_info,
        pipeline_name=pipeline_name,
        dagster_event=DagsterEvent(DagsterEventType.PIPELINE_FAILURE.value, pipeline_name),
    )


def build_process_start_event(run_id, pipeline_name):
    message = 'About to start process for pipeline {pipeline_name} run_id {run_id}'.format(
        pipeline_name=pipeline_name, run_id=run_id
    )

    return DagsterEventRecord(
        message=message,
        user_message=message,
        level=logging.INFO,
        run_id=run_id,
        timestamp=time.time(),
        error_info=None,
        pipeline_name=pipeline_name,
        dagster_event=DagsterEvent(DagsterEventType.PIPELINE_PROCESS_START.value, pipeline_name),
    )


def build_process_started_event(run_id, pipeline_name, process_id):
    message = 'Started process {process_id} for pipeline {pipeline_name} run_id {run_id}'.format(
        pipeline_name=pipeline_name, run_id=run_id, process_id=process_id
    )

    return DagsterEventRecord(
        message=message,
        user_message=message,
        level=logging.INFO,
        run_id=run_id,
        timestamp=time.time(),
        error_info=None,
        pipeline_name=pipeline_name,
        dagster_event=DagsterEvent(
            event_type_value=DagsterEventType.PIPELINE_PROCESS_STARTED.value,
            pipeline_name=pipeline_name,
            step_key=None,
            solid_handle=None,
            step_kind_value=None,
            logging_tags=None,
            event_specific_data=PipelineProcessStartedData(process_id),
        ),
    )


class SynchronousExecutionManager(PipelineExecutionManager):
    def execute_pipeline(self, _, pipeline, pipeline_run, raise_on_error):
        check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        try:
            return execute_pipeline(
                pipeline,
                pipeline_run.config,
                run_config=RunConfig(
                    pipeline_run.run_id,
                    mode=pipeline_run.mode,
                    event_callback=pipeline_run.handle_new_event,
                    executor_config=InProcessExecutorConfig(raise_on_error=raise_on_error),
                    reexecution_config=pipeline_run.reexecution_config,
                    step_keys_to_execute=pipeline_run.step_keys_to_execute,
                ),
            )
        except Exception:  # pylint: disable=broad-except
            if raise_on_error:
                six.reraise(*sys.exc_info())

            pipeline_run.handle_new_event(
                build_synthetic_pipeline_error_record(
                    pipeline_run.run_id,
                    serializable_error_info_from_exc_info(sys.exc_info()),
                    pipeline.name,
                )
            )


class MultiprocessingDone(object):
    pass


class MultiprocessingError(object):
    def __init__(self, error_info):
        self.error_info = check.inst_param(error_info, 'error_info', SerializableErrorInfo)


class ProcessStartedSentinel(object):
    def __init__(self, process_id):
        self.process_id = check.int_param(process_id, 'process_id')


class MultiprocessingExecutionManager(PipelineExecutionManager):
    def __init__(self):
        self._multiprocessing_context = get_multiprocessing_context()
        self._processes_lock = self._multiprocessing_context.Lock()
        self._processes = []
        # This is actually a reverse semaphore. We keep track of number of
        # processes we have by releasing semaphore every time we start
        # processing, we release after processing is finished
        self._processing_semaphore = gevent.lock.Semaphore(0)

        gevent.spawn(self._start_polling)
        atexit.register(self._cleanup)

    def _start_polling(self):
        while True:
            self._poll()
            gevent.sleep(0.1)

    def _cleanup(self):
        # Wait for child processes to finish and communicate on exit
        self.join()

    def _poll(self):
        with self._processes_lock:
            processes = copy.copy(self._processes)
            self._processes = []
            for _ in processes:
                self._processing_semaphore.release()

        for process in processes:
            done = self._consume_process_queue(process)
            if not done and not process.process.is_alive():
                done = self._consume_process_queue(process)
                if not done:
                    try:
                        done = True
                        raise Exception(
                            'Pipeline execution process for {run_id} unexpectedly exited'.format(
                                run_id=process.pipeline_run.run_id
                            )
                        )
                    except Exception:  # pylint: disable=broad-except
                        process.pipeline_run.handle_new_event(
                            build_synthetic_pipeline_error_record(
                                process.pipeline_run.run_id,
                                serializable_error_info_from_exc_info(sys.exc_info()),
                                process.pipeline_run.pipeline_name,
                            )
                        )

            if not done:
                with self._processes_lock:
                    self._processes.append(process)

            with self._processes_lock:
                self._processing_semaphore.acquire()

    def _consume_process_queue(self, process):
        while not process.message_queue.empty():
            message = process.message_queue.get(False)
            if isinstance(message, MultiprocessingDone):
                return True
            elif isinstance(message, MultiprocessingError):
                process.pipeline_run.handle_new_event(
                    build_synthetic_pipeline_error_record(
                        process.pipeline_run.run_id,
                        message.error_info,
                        process.pipeline_run.pipeline_name,
                    )
                )
            elif isinstance(message, ProcessStartedSentinel):
                process.pipeline_run.handle_new_event(
                    build_process_started_event(
                        process.pipeline_run.run_id,
                        process.pipeline_run.pipeline_name,
                        message.process_id,
                    )
                )
            else:
                process.pipeline_run.handle_new_event(message)
        return False

    def join(self):
        '''Waits until all there are no processes enqueued.'''
        while True:
            with self._processes_lock:
                if not self._processes and self._processing_semaphore.locked():
                    return True
            gevent.sleep(0.1)

    def execute_pipeline(self, handle, pipeline, pipeline_run, raise_on_error):
        check.inst_param(handle, 'handle', ExecutionTargetHandle)
        check.invariant(
            raise_on_error is False, 'Multiprocessing execute_pipeline does not rethrow user error'
        )

        message_queue = self._multiprocessing_context.Queue()
        p = self._multiprocessing_context.Process(
            target=execute_pipeline_through_queue,
            args=(
                handle,
                pipeline_run.selector.name,
                pipeline_run.selector.solid_subset,
                pipeline_run.config,
            ),
            kwargs={
                'run_id': pipeline_run.run_id,
                'mode': pipeline_run.mode,
                'message_queue': message_queue,
                'reexecution_config': pipeline_run.reexecution_config,
                'step_keys_to_execute': pipeline_run.step_keys_to_execute,
            },
        )

        pipeline_run.handle_new_event(build_process_start_event(pipeline_run.run_id, pipeline.name))

        p.start()
        with self._processes_lock:
            process = RunProcessWrapper(pipeline_run, p, message_queue)
            self._processes.append(process)


class RunProcessWrapper(namedtuple('RunProcessWrapper', 'pipeline_run process message_queue')):
    def __new__(cls, pipeline_run, process, message_queue):
        return super(RunProcessWrapper, cls).__new__(
            cls, check.inst_param(pipeline_run, 'pipeline_run', PipelineRun), process, message_queue
        )


def execute_pipeline_through_queue(
    handle,
    pipeline_name,
    solid_subset,
    environment_dict,
    mode,
    run_id,
    message_queue,
    reexecution_config,
    step_keys_to_execute,
):
    """
    Execute pipeline using message queue as a transport
    """

    check.opt_str_param(mode, 'mode')

    message_queue.put(ProcessStartedSentinel(os.getpid()))

    run_config = RunConfig(
        run_id,
        mode=mode,
        event_callback=message_queue.put,
        executor_config=InProcessExecutorConfig(raise_on_error=False),
        reexecution_config=reexecution_config,
        step_keys_to_execute=step_keys_to_execute,
    )

    try:
        handle.build_repository_definition()
        pipeline_def = handle.with_pipeline_name(pipeline_name).build_pipeline_definition()
    except Exception:  # pylint: disable=broad-except
        repo_error = sys.exc_info()
        message_queue.put(MultiprocessingError(serializable_error_info_from_exc_info(repo_error)))
        return

    try:
        result = execute_pipeline(
            pipeline_def.build_sub_pipeline(solid_subset), environment_dict, run_config=run_config
        )
        return result
    except Exception:  # pylint: disable=broad-except
        error_info = serializable_error_info_from_exc_info(sys.exc_info())
        message_queue.put(MultiprocessingError(error_info))
    finally:
        message_queue.put(MultiprocessingDone())
        message_queue.close()



from graphql.execution.base import ResolveInfo

from dagster import check, RunConfig
from dagster.core.definitions import create_environment_schema
from dagster.core.execution.api import ExecutionSelector, create_execution_plan
from dagster.core.types.evaluator import evaluate_config

from .fetch_pipelines import get_dauphin_pipeline_from_selector
from .utils import capture_dauphin_error, UserFacingGraphQLError


def validate_config(graphene_info, dauphin_pipeline, env_config, mode):
    get_validated_config(graphene_info, dauphin_pipeline, env_config, mode)


def get_validated_config(graphene_info, dauphin_pipeline, environment_dict, mode):
    check.opt_dict_param(environment_dict, 'environment_dict', key_type=str)
    check.str_param(mode, 'mode')

    pipeline = dauphin_pipeline.get_dagster_pipeline()

    environment_schema = create_environment_schema(pipeline, mode)

    validated_config = evaluate_config(
        environment_schema.environment_type, environment_dict, pipeline
    )

    if not validated_config.success:
        raise UserFacingGraphQLError(
            graphene_info.schema.type_named('PipelineConfigValidationInvalid')(
                pipeline=dauphin_pipeline,
                errors=[
                    graphene_info.schema.type_named(
                        'PipelineConfigValidationError'
                    ).from_dagster_error(graphene_info, err)
                    for err in validated_config.errors
                ],
            )
        )

    return validated_config


def get_run(graphene_info, runId):
    pipeline_run_storage = graphene_info.context.pipeline_runs
    run = pipeline_run_storage.get_run_by_id(runId)
    if not run:
        return graphene_info.schema.type_named('PipelineRunNotFoundError')(runId)
    else:
        return graphene_info.schema.type_named('PipelineRun')(run)


def get_runs(graphene_info):
    pipeline_run_storage = graphene_info.context.pipeline_runs
    return [
        graphene_info.schema.type_named('PipelineRun')(run)
        for run in pipeline_run_storage.all_runs()
    ]


@capture_dauphin_error
def validate_pipeline_config(graphene_info, selector, environment_dict, mode):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(selector, 'selector', ExecutionSelector)
    check.opt_str_param(mode, 'mode')

    dauphin_pipeline = get_dauphin_pipeline_from_selector(graphene_info, selector)
    get_validated_config(graphene_info, dauphin_pipeline, environment_dict, mode)
    return graphene_info.schema.type_named('PipelineConfigValidationValid')(dauphin_pipeline)


@capture_dauphin_error
def get_execution_plan(graphene_info, selector, environment_dict, mode):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(selector, 'selector', ExecutionSelector)
    check.opt_str_param(mode, 'mode')

    dauphin_pipeline = get_dauphin_pipeline_from_selector(graphene_info, selector)
    get_validated_config(graphene_info, dauphin_pipeline, environment_dict, mode)
    return graphene_info.schema.type_named('ExecutionPlan')(
        dauphin_pipeline,
        create_execution_plan(
            dauphin_pipeline.get_dagster_pipeline(), environment_dict, RunConfig(mode=mode)
        ),
    )



def capture_dauphin_error(fn):
    def _fn(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except UserFacingGraphQLError as de_exception:
            return de_exception.dauphin_error

    return _fn


class UserFacingGraphQLError(Exception):
    def __init__(self, dauphin_error, *args, **kwargs):
        self.dauphin_error = dauphin_error
        super(UserFacingGraphQLError, self).__init__(*args, **kwargs)



from dagster import check
from dagster.core.definitions import (
    CompositeSolidDefinition,
    IContainSolids,
    InputDefinition,
    ISolidDefinition,
    OutputDefinition,
    Solid,
    SolidDefinition,
    SolidHandle,
    SolidInputHandle,
    SolidOutputHandle,
)
from dagster_graphql import dauphin

from .runtime_types import to_dauphin_runtime_type


class DauphinSolidContainer(dauphin.Interface):
    class Meta:
        name = 'SolidContainer'

    solids = dauphin.non_null_list('Solid')


class DauphinISolidDefinition(dauphin.Interface):
    class Meta:
        name = 'ISolidDefinition'

    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    metadata = dauphin.non_null_list('MetadataItemDefinition')
    input_definitions = dauphin.non_null_list('InputDefinition')
    output_definitions = dauphin.non_null_list('OutputDefinition')


class ISolidDefinitionMixin:
    def __init__(self, solid_def):
        self._solid_def = check.inst_param(solid_def, 'solid_def', ISolidDefinition)

    def resolve_metadata(self, graphene_info):
        return [
            graphene_info.schema.type_named('MetadataItemDefinition')(key=item[0], value=item[1])
            for item in self._solid_def.metadata.items()
        ]

    def resolve_input_definitions(self, graphene_info):
        return [
            graphene_info.schema.type_named('InputDefinition')(input_definition, self)
            for input_definition in self._solid_def.input_defs
        ]

    def resolve_output_definitions(self, graphene_info):
        return [
            graphene_info.schema.type_named('OutputDefinition')(output_definition, self)
            for output_definition in self._solid_def.output_defs
        ]


class DauphinSolid(dauphin.ObjectType):
    class Meta:
        name = 'Solid'

    name = dauphin.NonNull(dauphin.String)
    definition = dauphin.NonNull('ISolidDefinition')
    inputs = dauphin.non_null_list('Input')
    outputs = dauphin.non_null_list('Output')

    def __init__(self, solid, depends_on=None, depended_by=None):
        super(DauphinSolid, self).__init__(name=solid.name)
        check.opt_dict_param(depends_on, 'depends_on', key_type=SolidInputHandle, value_type=list)
        check.opt_dict_param(
            depended_by, 'depended_by', key_type=SolidOutputHandle, value_type=list
        )

        self._solid = check.inst_param(solid, 'solid', Solid)

        if depends_on:
            self.depends_on = {
                input_handle: output_handles for input_handle, output_handles in depends_on.items()
            }
        else:
            self.depends_on = {}

        if depended_by:
            self.depended_by = {
                output_handle: input_handles for output_handle, input_handles in depended_by.items()
            }
        else:
            self.depended_by = {}

    def resolve_definition(self, graphene_info):
        if isinstance(self._solid.definition, SolidDefinition):
            return graphene_info.schema.type_named('SolidDefinition')(self._solid.definition)

        if isinstance(self._solid.definition, CompositeSolidDefinition):
            return graphene_info.schema.type_named('CompositeSolidDefinition')(
                self._solid.definition
            )

        check.invariant(
            False, 'Unknown solid definition type {type}'.format(type=type(self._solid.definition))
        )

    def resolve_inputs(self, graphene_info):
        return [
            graphene_info.schema.type_named('Input')(input_handle, self)
            for input_handle in self._solid.input_handles()
        ]

    def resolve_outputs(self, graphene_info):
        return [
            graphene_info.schema.type_named('Output')(output_handle, self)
            for output_handle in self._solid.output_handles()
        ]


class DauphinSolidDefinition(dauphin.ObjectType, ISolidDefinitionMixin):
    class Meta:
        name = 'SolidDefinition'
        interfaces = [DauphinISolidDefinition]

    config_definition = dauphin.Field('ConfigTypeField')

    def __init__(self, solid_def):
        check.inst_param(solid_def, 'solid_def', SolidDefinition)
        super(DauphinSolidDefinition, self).__init__(
            name=solid_def.name, description=solid_def.description
        )
        ISolidDefinitionMixin.__init__(self, solid_def)

    def resolve_config_definition(self, graphene_info):
        return (
            graphene_info.schema.type_named('ConfigTypeField')(
                name="config", field=self._solid_def.config_field
            )
            if self._solid_def.config_field
            else None
        )


class DauphinCompositeSolidDefinition(dauphin.ObjectType, ISolidDefinitionMixin):
    class Meta:
        name = 'CompositeSolidDefinition'
        interfaces = [DauphinISolidDefinition, DauphinSolidContainer]

    solids = dauphin.non_null_list('Solid')
    input_mappings = dauphin.non_null_list('InputMapping')
    output_mappings = dauphin.non_null_list('OutputMapping')

    def __init__(self, solid_def):
        check.inst_param(solid_def, 'solid_def', CompositeSolidDefinition)

        super(DauphinCompositeSolidDefinition, self).__init__(
            name=solid_def.name, description=solid_def.description
        )
        ISolidDefinitionMixin.__init__(self, solid_def)

    def resolve_solids(self, _graphene_info):
        return build_dauphin_solids(self._solid_def)

    def resolve_output_mappings(self, graphene_info):
        mappings = []
        for mapping in self._solid_def.output_mappings:
            mapped_solid = self._solid_def.solid_named(mapping.solid_name)
            mappings.append(
                graphene_info.schema.type_named('OutputMapping')(
                    graphene_info.schema.type_named('OutputDefinition')(
                        mapping.definition, self._solid_def
                    ),
                    graphene_info.schema.type_named('Output')(
                        mapped_solid.output_handle(mapping.output_name), DauphinSolid(mapped_solid)
                    ),
                )
            )
        return mappings

    def resolve_input_mappings(self, graphene_info):
        mappings = []
        for mapping in self._solid_def.input_mappings:
            mapped_solid = self._solid_def.solid_named(mapping.solid_name)
            mappings.append(
                graphene_info.schema.type_named('InputMapping')(
                    graphene_info.schema.type_named('InputDefinition')(
                        mapping.definition, self._solid_def
                    ),
                    graphene_info.schema.type_named('Input')(
                        mapped_solid.input_handle(mapping.input_name), DauphinSolid(mapped_solid)
                    ),
                )
            )
        return mappings


class DauphinSolidHandle(dauphin.ObjectType):
    class Meta:
        name = 'SolidHandle'

    handleID = dauphin.NonNull(dauphin.String)
    solid = dauphin.NonNull('Solid')
    parent = dauphin.Field('SolidHandle')

    def __init__(self, handle, solid, parent):
        self.handleID = check.inst_param(handle, 'handle', SolidHandle)
        self.solid = check.inst_param(solid, 'solid', DauphinSolid)
        self.parent = check.opt_inst_param(parent, 'parent', DauphinSolidHandle)


class DauphinInputDefinition(dauphin.ObjectType):
    class Meta:
        name = 'InputDefinition'

    solid_definition = dauphin.NonNull('SolidDefinition')
    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    type = dauphin.NonNull('RuntimeType')
    expectations = dauphin.non_null_list('Expectation')

    # inputs - ?

    def __init__(self, input_definition, solid_def):
        super(DauphinInputDefinition, self).__init__(
            name=input_definition.name,
            description=input_definition.description,
            solid_definition=solid_def,
        )
        self._input_definition = check.inst_param(
            input_definition, 'input_definition', InputDefinition
        )

    def resolve_type(self, _graphene_info):
        return to_dauphin_runtime_type(self._input_definition.runtime_type)

    def resolve_expectations(self, graphene_info):
        if self._input_definition.expectations:
            return [
                graphene_info.schema.type_named('Expectation')(expectation)
                for expectation in self._input_definition.expectations
            ]
        else:
            return []


class DauphinOutputDefinition(dauphin.ObjectType):
    class Meta:
        name = 'OutputDefinition'

    solid_definition = dauphin.NonNull('SolidDefinition')
    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    type = dauphin.NonNull('RuntimeType')
    expectations = dauphin.non_null_list('Expectation')

    # outputs - ?

    def __init__(self, output_definition, solid_def):
        super(DauphinOutputDefinition, self).__init__(
            name=output_definition.name,
            description=output_definition.description,
            solid_definition=solid_def,
        )
        self._output_definition = check.inst_param(
            output_definition, 'output_definition', OutputDefinition
        )

    def resolve_type(self, _graphene_info):
        return to_dauphin_runtime_type(self._output_definition.runtime_type)

    def resolve_expectations(self, graphene_info):
        if self._output_definition.expectations:
            return [
                graphene_info.schema.type_named('Expectation')(expectation)
                for expectation in self._output_definition.expectations
            ]
        else:
            return []


class DauphinInput(dauphin.ObjectType):
    class Meta:
        name = 'Input'

    solid = dauphin.NonNull('Solid')
    definition = dauphin.NonNull('InputDefinition')
    depends_on = dauphin.non_null_list('Output')

    def __init__(self, input_handle, solid):
        super(DauphinInput, self).__init__(solid=solid)
        self._solid = check.inst_param(solid, 'solid', DauphinSolid)
        self._input_handle = check.inst_param(input_handle, 'input_handle', SolidInputHandle)

    def resolve_definition(self, graphene_info):
        return graphene_info.schema.type_named('InputDefinition')(
            self._input_handle.input_def, self._solid.resolve_definition(graphene_info)
        )

    def resolve_depends_on(self, graphene_info):
        return [
            graphene_info.schema.type_named('Output')(
                dep, graphene_info.schema.type_named('Solid')(dep.solid)
            )
            for dep in self._solid.depends_on.get(self._input_handle, [])
        ]


class DauphinOutput(dauphin.ObjectType):
    class Meta:
        name = 'Output'

    solid = dauphin.NonNull('Solid')
    definition = dauphin.NonNull('OutputDefinition')
    depended_by = dauphin.non_null_list('Input')

    def __init__(self, output_handle, solid):
        super(DauphinOutput, self).__init__(solid=solid)
        self._solid = check.inst_param(solid, 'solid', DauphinSolid)
        self._output_handle = check.inst_param(output_handle, 'output_handle', SolidOutputHandle)

    def resolve_definition(self, graphene_info):
        return graphene_info.schema.type_named('OutputDefinition')(
            self._output_handle.output_def, self._solid.resolve_definition(graphene_info)
        )

    def resolve_depended_by(self, graphene_info):
        return [
            graphene_info.schema.type_named('Input')(input_handle, DauphinSolid(input_handle.solid))
            for input_handle in self._solid.depended_by.get(self._output_handle, [])
        ]


class DauphinInputMapping(dauphin.ObjectType):
    class Meta:
        name = 'InputMapping'

    mapped_input = dauphin.NonNull('Input')
    definition = dauphin.NonNull('InputDefinition')

    def __init__(self, input_definition, mapped_input):
        super(DauphinInputMapping, self).__init__()
        self.definition = check.inst_param(
            input_definition, 'intput_definition', DauphinInputDefinition
        )
        self.mapped_input = check.inst_param(mapped_input, 'mapped_input', DauphinInput)


class DauphinOutputMapping(dauphin.ObjectType):
    class Meta:
        name = 'OutputMapping'

    mapped_output = dauphin.NonNull('Output')
    definition = dauphin.NonNull('OutputDefinition')

    def __init__(self, output_definition, mapped_output):
        super(DauphinOutputMapping, self).__init__()
        self.definition = check.inst_param(
            output_definition, 'output_definition', DauphinOutputDefinition
        )
        self.mapped_output = check.inst_param(mapped_output, 'mapped_output', DauphinOutput)


def build_dauphin_solid(solid, deps):
    return DauphinSolid(
        solid, deps.deps_of_solid_with_input(solid.name), deps.depended_by_of_solid(solid.name)
    )


def build_dauphin_solids(container):
    check.inst_param(container, 'container', IContainSolids)
    return sorted(
        [build_dauphin_solid(solid, container.dependency_structure) for solid in container.solids],
        key=lambda solid: solid.name,
    )


def build_dauphin_solid_handles(container, parent=None):
    check.inst_param(container, 'container', IContainSolids)
    all_handle = []

    for solid in container.solids:
        handle = DauphinSolidHandle(
            solid=build_dauphin_solid(solid, container.dependency_structure),
            handle=SolidHandle(
                solid.name, solid.definition.name, parent.handleID if parent else None
            ),
            parent=parent if parent else None,
        )
        if isinstance(solid.definition, IContainSolids):
            all_handle += build_dauphin_solid_handles(solid.definition, handle)

        all_handle.append(handle)

    return all_handle



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_successful_one_part_execute_plan 1'] = {
    'executePlan': {
        '__typename': 'ExecutePlanSuccess',
        'hasFailures': False,
        'pipeline': {
            'name': 'csv_hello_world'
        },
        'stepEvents': [
            {
                '__typename': 'ExecutionStepStartEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepInputEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepOutputEvent',
                'outputName': 'result',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                },
                'valueRepr': "[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3)]), OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7)])]"
            },
            {
                '__typename': 'ExecutionStepSuccessEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            }
        ]
    }
}

snapshots['test_success_whole_execution_plan 1'] = {
    'executePlan': {
        '__typename': 'ExecutePlanSuccess',
        'hasFailures': False,
        'pipeline': {
            'name': 'csv_hello_world'
        },
        'stepEvents': [
            {
                '__typename': 'ExecutionStepStartEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepInputEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepOutputEvent',
                'outputName': 'result',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                },
                'valueRepr': "[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3)]), OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7)])]"
            },
            {
                '__typename': 'ExecutionStepSuccessEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepStartEvent',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepInputEvent',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepOutputEvent',
                'outputName': 'result',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                },
                'valueRepr': "[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3), ('sum_sq', 9)]), OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7), ('sum_sq', 49)])]"
            },
            {
                '__typename': 'ExecutionStepSuccessEvent',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                }
            }
        ]
    }
}

snapshots['test_success_whole_execution_plan_with_filesystem_config 1'] = {
    'executePlan': {
        '__typename': 'ExecutePlanSuccess',
        'hasFailures': False,
        'pipeline': {
            'name': 'csv_hello_world'
        },
        'stepEvents': [
            {
                '__typename': 'ExecutionStepStartEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepInputEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepOutputEvent',
                'outputName': 'result',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                },
                'valueRepr': "[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3)]), OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7)])]"
            },
            {
                '__typename': 'ExecutionStepSuccessEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepStartEvent',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepInputEvent',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepOutputEvent',
                'outputName': 'result',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                },
                'valueRepr': "[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3), ('sum_sq', 9)]), OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7), ('sum_sq', 49)])]"
            },
            {
                '__typename': 'ExecutionStepSuccessEvent',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                }
            }
        ]
    }
}

snapshots['test_success_whole_execution_plan_with_in_memory_config 1'] = {
    'executePlan': {
        '__typename': 'ExecutePlanSuccess',
        'hasFailures': False,
        'pipeline': {
            'name': 'csv_hello_world'
        },
        'stepEvents': [
            {
                '__typename': 'ExecutionStepStartEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepInputEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepOutputEvent',
                'outputName': 'result',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                },
                'valueRepr': "[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3)]), OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7)])]"
            },
            {
                '__typename': 'ExecutionStepSuccessEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepStartEvent',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepInputEvent',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepOutputEvent',
                'outputName': 'result',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                },
                'valueRepr': "[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3), ('sum_sq', 9)]), OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7), ('sum_sq', 49)])]"
            },
            {
                '__typename': 'ExecutionStepSuccessEvent',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                }
            }
        ]
    }
}

snapshots['test_successful_two_part_execute_plan 1'] = {
    'executePlan': {
        '__typename': 'ExecutePlanSuccess',
        'hasFailures': False,
        'pipeline': {
            'name': 'csv_hello_world'
        },
        'stepEvents': [
            {
                '__typename': 'ExecutionStepStartEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepInputEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepOutputEvent',
                'outputName': 'result',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                },
                'valueRepr': "[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3)]), OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7)])]"
            },
            {
                '__typename': 'ExecutionStepSuccessEvent',
                'step': {
                    'key': 'sum_solid.compute',
                    'metadata': [
                    ]
                }
            }
        ]
    }
}

snapshots['test_successful_two_part_execute_plan 2'] = {
    'executePlan': {
        '__typename': 'ExecutePlanSuccess',
        'hasFailures': False,
        'pipeline': {
            'name': 'csv_hello_world'
        },
        'stepEvents': [
            {
                '__typename': 'ExecutionStepStartEvent',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepInputEvent',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                }
            },
            {
                '__typename': 'ExecutionStepOutputEvent',
                'outputName': 'result',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                },
                'valueRepr': "[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3), ('sum_sq', 9)]), OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7), ('sum_sq', 49)])]"
            },
            {
                '__typename': 'ExecutionStepSuccessEvent',
                'step': {
                    'key': 'sum_sq_solid.compute',
                    'metadata': [
                    ]
                }
            }
        ]
    }
}

snapshots['test_invalid_config_fetch_execute_plan 1'] = {
    'executionPlan': {
        '__typename': 'PipelineConfigValidationInvalid',
        'errors': [
            {
                'message': 'Value at path root:solids:sum_solid:inputs:num is not valid. Expected "Path"'
            }
        ],
        'pipeline': {
            'name': 'csv_hello_world'
        }
    }
}

snapshots['test_invalid_config_execute_plan 1'] = {
    'executePlan': {
        '__typename': 'PipelineConfigValidationInvalid',
        'errors': [
            {
                'message': 'Value at path root:solids:sum_solid:inputs:num is not valid. Expected "Path"'
            }
        ],
        'pipeline': {
            'name': 'csv_hello_world'
        }
    }
}

snapshots['test_pipeline_not_found_error_execute_plan 1'] = {
    'executePlan': {
        '__typename': 'PipelineNotFoundError',
        'pipelineName': 'nope'
    }
}

snapshots['test_pipeline_with_execution_metadata 1'] = {
    'executionPlan': {
        '__typename': 'ExecutionPlan',
        'pipeline': {
            'name': 'pipeline_with_step_metadata'
        },
        'steps': [
            {
                'inputs': [
                ],
                'key': 'solid_metadata_creation.compute',
                'kind': 'COMPUTE',
                'metadata': [
                    {
                        'key': 'computed',
                        'value': 'foobar1'
                    }
                ],
                'outputs': [
                ],
                'solidHandleID': 'solid_metadata_creation'
            }
        ]
    }
}



from dagster import check
from dagster.core.types.config import ConfigType
from dagster.core.types.field_utils import FieldImpl

from dagster_graphql import dauphin


def to_dauphin_config_type(config_type):
    check.inst_param(config_type, 'config_type', ConfigType)

    if config_type.is_enum:
        return DauphinEnumConfigType(config_type)
    elif config_type.has_fields:
        return DauphinCompositeConfigType(config_type)
    elif config_type.is_list:
        return DauphinListConfigType(config_type)
    elif config_type.is_nullable:
        return DauphinNullableConfigType(config_type)
    else:
        return DauphinRegularConfigType(config_type)


def _ctor_kwargs(config_type):
    return dict(
        key=config_type.key,
        name=config_type.name,
        description=config_type.description,
        is_builtin=config_type.type_attributes.is_builtin,
        is_list=config_type.is_list,
        is_nullable=config_type.is_nullable,
        is_selector=config_type.is_selector,
        is_system_generated=config_type.type_attributes.is_system_config,
    )


class DauphinConfigType(dauphin.Interface):
    class Meta:
        name = 'ConfigType'

    key = dauphin.NonNull(dauphin.String)
    name = dauphin.String()
    description = dauphin.String()

    inner_types = dauphin.non_null_list('ConfigType')

    is_nullable = dauphin.NonNull(dauphin.Boolean)
    is_list = dauphin.NonNull(dauphin.Boolean)
    is_selector = dauphin.NonNull(dauphin.Boolean)

    is_builtin = dauphin.NonNull(
        dauphin.Boolean,
        description='''
True if the system defines it and it is the same type across pipelines.
Examples include "Int" and "String."''',
    )

    is_system_generated = dauphin.NonNull(
        dauphin.Boolean,
        description='''
Dagster generates types for base elements of the config system (e.g. the solids and
context field of the base environment). These types are always present
and are typically not relevant to an end user. This flag allows tool authors to
filter out those types by default.
''',
    )


def _resolve_inner_types(config_type):
    return list(map(to_dauphin_config_type, config_type.inner_types))


class DauphinRegularConfigType(dauphin.ObjectType):
    def __init__(self, config_type):
        self._config_type = check.inst_param(config_type, 'config_type', ConfigType)
        super(DauphinRegularConfigType, self).__init__(**_ctor_kwargs(config_type))

    class Meta:
        name = 'RegularConfigType'
        interfaces = [DauphinConfigType]

    def resolve_inner_types(self, _graphene_info):
        return _resolve_inner_types(self._config_type)


class DauphinWrappingConfigType(dauphin.Interface):
    class Meta:
        name = 'WrappingConfigType'

    of_type = dauphin.Field(dauphin.NonNull(DauphinConfigType))


class DauphinListConfigType(dauphin.ObjectType):
    def __init__(self, config_type):
        self._config_type = check.inst_param(config_type, 'config_type', ConfigType)
        super(DauphinListConfigType, self).__init__(**_ctor_kwargs(config_type))

    class Meta:
        name = 'ListConfigType'
        interfaces = [DauphinConfigType, DauphinWrappingConfigType]

    def resolve_of_type(self, _graphene_info):
        return to_dauphin_config_type(self._config_type.inner_type)

    def resolve_inner_types(self, _graphene_info):
        return _resolve_inner_types(self._config_type)


class DauphinNullableConfigType(dauphin.ObjectType):
    def __init__(self, config_type):
        self._config_type = check.inst_param(config_type, 'config_type', ConfigType)
        super(DauphinNullableConfigType, self).__init__(**_ctor_kwargs(config_type))

    class Meta:
        name = 'NullableConfigType'
        interfaces = [DauphinConfigType, DauphinWrappingConfigType]

    def resolve_of_type(self, _graphene_info):
        return to_dauphin_config_type(self._config_type.inner_type)

    def resolve_inner_types(self, _graphene_info):
        return _resolve_inner_types(self._config_type)


class DauphinEnumConfigType(dauphin.ObjectType):
    def __init__(self, config_type):
        check.inst_param(config_type, 'config_type', ConfigType)
        check.param_invariant(config_type.is_enum, 'config_type')
        self._config_type = config_type
        super(DauphinEnumConfigType, self).__init__(**_ctor_kwargs(config_type))

    class Meta:
        name = 'EnumConfigType'
        interfaces = [DauphinConfigType]

    values = dauphin.non_null_list('EnumConfigValue')

    def resolve_values(self, _graphene_info):
        return [
            DauphinEnumConfigValue(value=ev.config_value, description=ev.description)
            for ev in self._config_type.enum_values
        ]

    def resolve_inner_types(self, _graphene_info):
        return _resolve_inner_types(self._config_type)


class DauphinEnumConfigValue(dauphin.ObjectType):
    class Meta:
        name = 'EnumConfigValue'

    value = dauphin.NonNull(dauphin.String)
    description = dauphin.String()


class DauphinCompositeConfigType(dauphin.ObjectType):
    def __init__(self, config_type):
        check.inst_param(config_type, 'config_type', ConfigType)
        check.param_invariant(config_type.has_fields, 'config_type')
        self._config_type = config_type
        super(DauphinCompositeConfigType, self).__init__(**_ctor_kwargs(config_type))

    class Meta:
        name = 'CompositeConfigType'
        interfaces = [DauphinConfigType]

    fields = dauphin.non_null_list('ConfigTypeField')

    def resolve_fields(self, _graphene_info):
        return sorted(
            [
                DauphinConfigTypeField(name=name, field=field)
                for name, field in self._config_type.fields.items()
            ],
            key=lambda field: field.name,
        )

    def resolve_inner_types(self, _graphene_info):
        return _resolve_inner_types(self._config_type)


class DauphinConfigTypeField(dauphin.ObjectType):
    class Meta:
        name = 'ConfigTypeField'

    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    config_type = dauphin.NonNull('ConfigType')
    default_value = dauphin.String()
    is_optional = dauphin.NonNull(dauphin.Boolean)
    is_secret = dauphin.NonNull(dauphin.Boolean)

    def __init__(self, name, field):
        check.str_param(name, 'name')
        check.inst_param(field, 'field', FieldImpl)

        super(DauphinConfigTypeField, self).__init__(
            name=name,
            description=field.description,
            default_value=field.default_value_as_str if field.default_provided else None,
            is_optional=field.is_optional,
            is_secret=field.is_secret,
        )
        self._field = field

    def resolve_config_type(self, _graphene_info):
        return to_dauphin_config_type(self._field.config_type)



from dagster import check, ExecutionTargetHandle
from .pipeline_run_storage import PipelineRunStorage
from .pipeline_execution_manager import PipelineExecutionManager


class DagsterGraphQLContext(object):
    def __init__(
        self, handle, pipeline_runs, execution_manager, raise_on_error=False, version=None
    ):
        self._handle = check.inst_param(handle, 'handle', ExecutionTargetHandle)
        self.pipeline_runs = check.inst_param(pipeline_runs, 'pipeline_runs', PipelineRunStorage)
        self.execution_manager = check.inst_param(
            execution_manager, 'pipeline_execution_manager', PipelineExecutionManager
        )
        self.raise_on_error = check.bool_param(raise_on_error, 'raise_on_error')
        self.version = version

    def get_handle(self):
        return self._handle

    def get_repository(self):
        return self.get_handle().build_repository_definition()

    def get_pipeline(self, pipeline_name):
        orig_handle = self.get_handle()
        if orig_handle.is_resolved_to_pipeline:
            pipeline_def = orig_handle.build_pipeline_definition()
            check.invariant(
                pipeline_def.name == pipeline_name,
                '''Dagster GraphQL Context resolved pipeline with name {handle_pipeline_name},
                couldn't resolve {pipeline_name}'''.format(
                    handle_pipeline_name=pipeline_def.name, pipeline_name=pipeline_name
                ),
            )
            return pipeline_def
        return self.get_handle().with_pipeline_name(pipeline_name).build_pipeline_definition()



import sys

from graphql.execution.base import ResolveInfo

from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.core.execution.api import ExecutionSelector
from dagster.utils.error import serializable_error_info_from_exc_info

from .utils import UserFacingGraphQLError, capture_dauphin_error


@capture_dauphin_error
def get_pipeline_or_error(graphene_info, selector):
    return get_dauphin_pipeline_from_selector(graphene_info, selector)


def get_pipeline_or_raise(graphene_info, selector):
    return get_dauphin_pipeline_from_selector(graphene_info, selector)


@capture_dauphin_error
def get_pipelines_or_error(graphene_info):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    return _get_pipelines(graphene_info)


def get_pipelines_or_raise(graphene_info):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    return _get_pipelines(graphene_info)


def _get_pipelines(graphene_info):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)

    repository = graphene_info.context.get_repository()

    try:
        pipeline_instances = []
        for pipeline_def in repository.get_all_pipelines():
            pipeline_instances.append(graphene_info.schema.type_named('Pipeline')(pipeline_def))
        return graphene_info.schema.type_named('PipelineConnection')(
            nodes=sorted(pipeline_instances, key=lambda pipeline: pipeline.name)
        )
    except DagsterInvalidDefinitionError:
        raise UserFacingGraphQLError(
            graphene_info.schema.type_named('InvalidDefinitionError')(
                serializable_error_info_from_exc_info(sys.exc_info())
            )
        )
    except Exception:  # pylint: disable=broad-except
        raise UserFacingGraphQLError(
            graphene_info.schema.type_named('PythonError')(
                serializable_error_info_from_exc_info(sys.exc_info())
            )
        )


def get_dagster_pipeline_from_selector(graphene_info, selector):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(selector, 'selector', ExecutionSelector)
    repository = graphene_info.context.get_repository()
    if not repository.has_pipeline(selector.name):
        raise UserFacingGraphQLError(
            graphene_info.schema.type_named('PipelineNotFoundError')(pipeline_name=selector.name)
        )

    orig_pipeline = graphene_info.context.get_pipeline(selector.name)
    if selector.solid_subset:
        for solid_name in selector.solid_subset:
            if not orig_pipeline.has_solid_named(solid_name):
                raise UserFacingGraphQLError(
                    graphene_info.schema.type_named('SolidNotFoundError')(solid_name=solid_name)
                )
    return orig_pipeline.build_sub_pipeline(selector.solid_subset)


def get_dauphin_pipeline_from_selector(graphene_info, selector):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(selector, 'selector', ExecutionSelector)
    return graphene_info.schema.type_named('Pipeline')(
        get_dagster_pipeline_from_selector(graphene_info, selector)
    )



import json

from collections import namedtuple

from dagster import check

from dagster.core.definitions import ExpectationResult, Materialization, SolidHandle

from dagster.core.events import (
    DagsterEvent,
    DagsterEventType,
    StepExpectationResultData,
    StepMaterializationData,
)
from dagster.core.execution.plan.objects import (
    StepFailureData,
    StepInputData,
    StepOutputData,
    StepOutputHandle,
    StepSuccessData,
    TypeCheckData,
    UserFailureData,
)
from dagster.utils.error import SerializableErrorInfo

# Fragment for exhaustively retrieving all properties of step events
GraphQLFragment = namedtuple('_StepEventFragment', 'include_key fragment')


def get_step_event_fragment():
    return GraphQLFragment(
        include_key='...stepEventFragment',
        fragment='''
fragment eventMetadataEntryFragment on EventMetadataEntry {
  __typename
  label
  description
  ... on EventPathMetadataEntry {
      path
  }
  ... on EventJsonMetadataEntry {
      jsonString
  }
}

fragment stepEventFragment on StepEvent {
  step {
    key
    inputs {
      name
      type {
        key
      }
      dependsOn {
        key
      }
    }
    outputs {
      name
      type {
        key
      }
    }
    solidHandleID
    kind
    metadata {
      key
      value
    }
  }
  ... on MessageEvent {
    runId
    message
    timestamp
    level
  }
  ... on StepExpectationResultEvent {
    expectationResult {
      success
      label
      description
      metadataEntries {
        ...eventMetadataEntryFragment
      }
    }
  }
  ... on StepMaterializationEvent {
    materialization {
      label
      description
      metadataEntries {
        ...eventMetadataEntryFragment
      }
    }
  }
  ... on ExecutionStepInputEvent {
    inputName
    valueRepr
    typeCheck {
      __typename
      success
      label
      description
      metadataEntries {
        ...eventMetadataEntryFragment
      }
    }
  }
  ... on ExecutionStepOutputEvent {
    outputName
    valueRepr
    typeCheck {
      __typename
      success
      label
      description
      metadataEntries {
        ...eventMetadataEntryFragment
      }
    }
  }
  ... on ExecutionStepFailureEvent {
    error {
      message
    }
    failureMetadata {
      label
      description
      metadataEntries {
        ...eventMetadataEntryFragment
      }
    }
  }
}
'''.strip(),
    )


def get_log_message_event_fragment():
    return GraphQLFragment(
        include_key='...logMessageEventFragment',
        fragment='''
fragment logMessageEventFragment on LogMessageEvent {
  runId
  message
  timestamp
  level
  step {
    key
    inputs {
      name
      type {
        key
      }
      dependsOn {
        key
      }
    }
    outputs {
      name
      type {
        key
      }
    }
    solidHandleID
    kind
    metadata {
      key
      value
    }
  }
}
  ''',
    )


def _handled_events():
    return {
        'ExecutionStepStartEvent': DagsterEventType.STEP_START,
        'ExecutionStepInputEvent': DagsterEventType.STEP_INPUT,
        'ExecutionStepOutputEvent': DagsterEventType.STEP_OUTPUT,
        'ExecutionStepFailureEvent': DagsterEventType.STEP_FAILURE,
        'ExecutionStepSkippedEvent': DagsterEventType.STEP_SKIPPED,
        'ExecutionStepSuccessEvent': DagsterEventType.STEP_SUCCESS,
        'StepMaterializationEvent': DagsterEventType.STEP_MATERIALIZATION,
        'StepExpectationResultEvent': DagsterEventType.STEP_EXPECTATION_RESULT,
    }


from dagster import EventMetadataEntry


def expectation_result_from_data(data):
    return ExpectationResult(
        success=data['success'],
        label=data['label'],
        description=data.get('description'),  # enforce?
        metadata_entries=list(event_metadata_entries(data.get('metadataEntries')) or []),
    )


def materialization_from_data(data):
    return Materialization(
        label=data['label'],
        description=data.get('description'),  # enforce?
        metadata_entries=list(event_metadata_entries(data.get('metadataEntries')) or []),
    )


def event_metadata_entries(metadata_entry_datas):
    if not metadata_entry_datas:
        return

    for metadata_entry_data in metadata_entry_datas:
        typename = metadata_entry_data['__typename']
        label = metadata_entry_data['label']
        description = metadata_entry_data.get('description')
        if typename == 'EventPathMetadataEntry':
            yield EventMetadataEntry.path(
                label=label, description=description, path=metadata_entry_data['path']
            )
        elif typename == 'EventJsonMetadataEntry':
            yield EventMetadataEntry.json(
                label=label,
                description=description,
                data=json.loads(metadata_entry_data.get('jsonString', '')),
            )
        else:
            check.not_implemented('TODO for type {}'.format(typename))


def dagster_event_from_dict(event_dict, pipeline_name):
    check.dict_param(event_dict, 'event_dict', key_type=str)
    check.str_param(pipeline_name, 'pipeline_name')

    # Get event_type
    event_type = _handled_events().get(event_dict['__typename'])
    if not event_type:
        raise Exception('unhandled event type %s' % event_dict['__typename'])

    # Get event_specific_data
    event_specific_data = None
    if event_type == DagsterEventType.STEP_OUTPUT:
        event_specific_data = StepOutputData(
            step_output_handle=StepOutputHandle(
                event_dict['step']['key'], event_dict['outputName']
            ),
            value_repr=event_dict['valueRepr'],
            type_check_data=TypeCheckData(
                success=event_dict['typeCheck']['success'],
                label=event_dict['typeCheck']['label'],
                description=event_dict.get('description'),
                metadata_entries=list(
                    event_metadata_entries(event_dict.get('metadataEntries')) or []
                ),
            ),
        )

    elif event_type == DagsterEventType.STEP_INPUT:
        event_specific_data = StepInputData(
            input_name=event_dict['inputName'],
            value_repr=event_dict['valueRepr'],
            type_check_data=TypeCheckData(
                success=event_dict['typeCheck']['success'],
                label=event_dict['typeCheck']['label'],
                description=event_dict.get('description'),
                metadata_entries=list(
                    event_metadata_entries(event_dict.get('metadataEntries')) or []
                ),
            ),
        )
    elif event_type == DagsterEventType.STEP_SUCCESS:
        event_specific_data = StepSuccessData(0.0)

    elif event_type == DagsterEventType.STEP_MATERIALIZATION:
        materialization = event_dict['materialization']
        event_specific_data = StepMaterializationData(
            materialization=materialization_from_data(materialization)
        )
    elif event_type == DagsterEventType.STEP_EXPECTATION_RESULT:
        expectation_result = expectation_result_from_data(event_dict['expectationResult'])
        event_specific_data = StepExpectationResultData(expectation_result)

    elif event_type == DagsterEventType.STEP_FAILURE:
        error_info = SerializableErrorInfo(
            event_dict['error']['message'], stack=None, cls_name=None
        )
        event_specific_data = StepFailureData(
            error_info,
            UserFailureData(
                label=event_dict['failureMetadata']['label'],
                description=event_dict['failureMetadata']['description'],
                metadata_entries=list(
                    event_metadata_entries(event_dict.get('metadataEntries')) or []
                ),
            )
            if event_dict.get('failureMetadata')
            else None,
        )

    return DagsterEvent(
        event_type_value=event_type.value,
        pipeline_name=pipeline_name,
        step_key=event_dict['step']['key'],
        solid_handle=SolidHandle(event_dict['step']['solidHandleID'], None, None),
        step_kind_value=event_dict['step']['kind'],
        logging_tags=None,
        event_specific_data=event_specific_data,
    )



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_pipeline_reexecution_info_query 1'] = [
    'sum_sq_solid.compute'
]

snapshots['test_successful_pipeline_reexecution 1'] = {
    'startPipelineExecution': {
        '__typename': 'StartPipelineExecutionSuccess',
        'run': {
            'logs': {
                'nodes': [
                    {
                        '__typename': 'PipelineStartEvent',
                        'level': 'DEBUG'
                    },
                    {
                        '__typename': 'ExecutionStepStartEvent',
                        'level': 'DEBUG',
                        'step': {
                            'kind': 'COMPUTE'
                        }
                    },
                    {
                        '__typename': 'ExecutionStepInputEvent',
                        'inputName': 'num',
                        'level': 'DEBUG',
                        'step': {
                            'key': 'sum_solid.compute',
                            'kind': 'COMPUTE'
                        },
                        'typeCheck': {
                            'description': None,
                            'label': 'num',
                            'metadataEntries': [
                            ]
                        }
                    },
                    {
                        '__typename': 'LogMessageEvent',
                        'level': 'INFO'
                    },
                    {
                        '__typename': 'ExecutionStepOutputEvent',
                        'level': 'DEBUG',
                        'outputName': 'result',
                        'step': {
                            'key': 'sum_solid.compute',
                            'kind': 'COMPUTE'
                        },
                        'typeCheck': {
                            'description': None,
                            'label': 'result',
                            'metadataEntries': [
                            ]
                        }
                    },
                    {
                        '__typename': 'ExecutionStepSuccessEvent',
                        'level': 'DEBUG'
                    },
                    {
                        '__typename': 'ExecutionStepStartEvent',
                        'level': 'DEBUG',
                        'step': {
                            'kind': 'COMPUTE'
                        }
                    },
                    {
                        '__typename': 'ExecutionStepInputEvent',
                        'inputName': 'sum_df',
                        'level': 'DEBUG',
                        'step': {
                            'key': 'sum_sq_solid.compute',
                            'kind': 'COMPUTE'
                        },
                        'typeCheck': {
                            'description': None,
                            'label': 'sum_df',
                            'metadataEntries': [
                            ]
                        }
                    },
                    {
                        '__typename': 'LogMessageEvent',
                        'level': 'INFO'
                    },
                    {
                        '__typename': 'ExecutionStepOutputEvent',
                        'level': 'DEBUG',
                        'outputName': 'result',
                        'step': {
                            'key': 'sum_sq_solid.compute',
                            'kind': 'COMPUTE'
                        },
                        'typeCheck': {
                            'description': None,
                            'label': 'result',
                            'metadataEntries': [
                            ]
                        }
                    },
                    {
                        '__typename': 'ExecutionStepSuccessEvent',
                        'level': 'DEBUG'
                    },
                    {
                        '__typename': 'PipelineSuccessEvent',
                        'level': 'DEBUG'
                    }
                ]
            },
            'pipeline': {
                'name': 'csv_hello_world'
            }
        }
    }
}

snapshots['test_successful_pipeline_reexecution 2'] = {
    'startPipelineExecution': {
        '__typename': 'StartPipelineExecutionSuccess',
        'run': {
            'logs': {
                'nodes': [
                    {
                        '__typename': 'PipelineStartEvent',
                        'level': 'DEBUG'
                    },
                    {
                        '__typename': 'ExecutionStepStartEvent',
                        'level': 'DEBUG',
                        'step': {
                            'kind': 'COMPUTE'
                        }
                    },
                    {
                        '__typename': 'ExecutionStepInputEvent',
                        'inputName': 'sum_df',
                        'level': 'DEBUG',
                        'step': {
                            'key': 'sum_sq_solid.compute',
                            'kind': 'COMPUTE'
                        },
                        'typeCheck': {
                            'description': None,
                            'label': 'sum_df',
                            'metadataEntries': [
                            ]
                        }
                    },
                    {
                        '__typename': 'LogMessageEvent',
                        'level': 'INFO'
                    },
                    {
                        '__typename': 'ExecutionStepOutputEvent',
                        'level': 'DEBUG',
                        'outputName': 'result',
                        'step': {
                            'key': 'sum_sq_solid.compute',
                            'kind': 'COMPUTE'
                        },
                        'typeCheck': {
                            'description': None,
                            'label': 'result',
                            'metadataEntries': [
                            ]
                        }
                    },
                    {
                        '__typename': 'ExecutionStepSuccessEvent',
                        'level': 'DEBUG'
                    },
                    {
                        '__typename': 'PipelineSuccessEvent',
                        'level': 'DEBUG'
                    }
                ]
            },
            'pipeline': {
                'name': 'csv_hello_world'
            }
        }
    }
}



from dagster_graphql.test.preset_query import execute_preset_query
from .setup import define_context


def test_basic_preset_query_no_presets():
    result = execute_preset_query('csv_hello_world_two', define_context())
    assert result.data == {'pipeline': {'name': 'csv_hello_world_two', 'presets': []}}


def test_basic_preset_query_with_presets(snapshot):
    result = execute_preset_query('csv_hello_world', define_context())

    assert [preset_data['name'] for preset_data in result.data['pipeline']['presets']] == [
        'prod',
        'test',
    ]

    snapshot.assert_match(result.data)
